/*
 *	File:	Koliada - event.h
 *
 *	Contains: Koliada EVENT header
 *
 * Edition: TINY
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Copyright  1989-2015 by Guy McIlroy
 *  All rights reserved.
 *
 *  This module contains confidential, unpublished, proprietary source code.
 *  The copyright notice above does not evidence any actual or intended
 *  publication of such source code.
 *
 *  This code may only be used under licence from Koliada, LLC - www.koliada.com
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */
#ifndef __EVENT_H
#define __EVENT_H

#include "sysPower.h"

#include "object.h"

#define kEventType "event"
#ifndef __EVENT_STRUCT_H
	// external
	typedef class event;
	#define kEventClass ((event)kEventType)
#else
	// internal (avoid name collision)
	typedef class_type event_type;
	#define kEventClass ((event_type)kEventType)
#endif
#ifndef kMaxEventArgs
	#define kMaxEventArgs              4   // words
#endif
#define kEventObjectSize (kObjectSize + 1 + kMaxEventArgs)	// words (>excluding< object bit struct)

#define DefineEvent(e) DefineObject(EVENT, kEventType, kEventObjectSize, e)
#define ExternEvent(e) ExternObject(EVENT, e)

///////////////////////////////////////////////////////////////////////////////////////
//
// event inheritance info
//
///////////////////////////////////////////////////////////////////////////////////////

#define EVENT_HEADER \
	word __event_private_data[kEventObjectSize - kObjectSize];	/* private - excludes BITS */

#define EVENT_BITS	\
	OBJECT_BITS \
	/* private */ word _event_bits :3; \
	/* public  */ word fired       :1; \
	/* public  */ word live        :1; \
	/*         */

// event API indices
enum
	{
	__PostEvent = __objectMaxApiIndex,
	__OnEvent,
	__clearEvent,

	__eventMaxApiIndex,
	};

// event Alerts
enum
	{
	kEventErrorEventQueueFull=1, // */ {kError, "event queue is full"},
	kEventErrorBadIRQ,           // */ {kError, "IRQ %u is out of range"},
	kEventErrorAlreadyPosted,    // */ {kDebug, "event already posted"},
	kEventErrorBadPriority,      // */ {kError, "%u: bad event priority"},
	//
	kMaxEventError
	};

///////////////////////////////////////////////////////////////////////////////////////
//
// event API
//
///////////////////////////////////////////////////////////////////////////////////////

// IRQ handling
// in general, interrupts are handled using events
// for lowest latency, use a dedicated interrupt handler (see interrupt.h)
// events are generally handled FIFO, but may be prioritized as needed
// event prioritization is implementation specific
//
// measured event latencies (from interrupt handler to event handler) are;
//   - 8051         ????us
//   - MSP430       ????us
//   - LM(CortexM3) ????us
// realized latency will depend on load and event prioritization
//
// actual interrupt vectors move around depending on the SoC so
// event maps system vectors for dispatch via sys->ITBL[]
//
// sys->ITBL[] event indices are target specific (see target.h)
//
// use;
//		objectCreate(event);
// to create an event
//

// events may be separated by the queue they are posted to
typedef enum
	{
	kTrapEvent,	// real-time 'trap' events (unqueued)
//	kFastEvent,	// fast user events (LIFO)
	kUserEvent,	// normal user events (FIFO)
	//
	kMaxEventQueues
	} EventPriority;

// PostIRQ posts an event in the IRQ table
#define PostIRQ(isr, ...)	do{ register EVENT obj=_sys->ITBL[isr##_ISR]; if (obj) PostEvent(obj, __VA_ARGS__); }while(0)

STATUS OnEvent(EVENT, HANDLER);				// connect a handler to an event
STATUS PostEvent(EVENT, ...);					// post an event as fired
														// PostEvent(e, arg1, arg2...);
STATUS clearEvent(EVENT);						// clear an event

//
// statics
//
// all events have the time they were last fired as an event attribute;
//TickCount getEventTime(EVENT);                     // time (ticks) event last fired
//TickCount getEventDelta(EVENT);                    // time delta (ticks) between last and previous fires

STATUS WaitEvent(EVENT, word, ...);			// wait for event to fire
																		// WaitEvent(NULL|ANY, 0);
																		// WaitEvent(e, argc, &arg1, &arg2...);
#define eventEnable(e, b) objectEnable((OBJECT) e, b)
#define eventTrace(e, b) objectTrace((OBJECT) e, b)

STATUS EventDefer();	// defer to higher priority events
STATUS EventYield();	// yield to any other events

STATUS setEventIRQ(word irq, EVENT obj);
OBJECT getEventIRQ(word irq);

OBJECT setEventPriority(EVENT obj, word priority);
OBJECT getEventPriority(EVENT obj);

word   setEventSleepMode(word mode);
word   getEventSleepMode();

SysPowerMode (*getEventSleepNotice())(SysPowerMode);
SysPowerMode (*setEventSleepNotice(SysPowerMode (*notify)(SysPowerMode)))(SysPowerMode);

void (*setEventAwakeNotice(SysPowerMode (*notify)(SysPowerMode)))(SysPowerMode);
void (*getEventAwakeNotice())(SysPowerMode);

#endif//__EVENT_H

