/*
 *	File:	Koliada - object.h
 *
 *	Contains: KoliadaES OBJECT header
 *
 * Edition: TINY
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Copyright  1989-2015 by Guy McIlroy
 *  All rights reserved.
 *
 *  This module contains confidential, unpublished, proprietary source code.
 *  The copyright notice above does not evidence any actual or intended
 *  publication of such source code.
 *
 *  This code may only be used under licence from Koliada, LLC - www.koliada.com
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */
#ifndef __OBJECT_H
#define __OBJECT_H

typedef unsigned int ObjectSize;
typedef const char *ObjectName;
typedef const char *ObjectType;
typedef const char *ObjectClass;

#define nullObject ((OBJECT) 0)

// TE - objects are defined locally

#define kObjectType "object"
#ifndef __OBJECT_STRUCT_H
	// external
	typedef word object;
	typedef object class;
	#define kObjectClass ((object)kObjectType)
#else
	// internal (avoid name collision)
	typedef word object_type;
	typedef word class_type;
	#define kObjectClass ((object_type)kObjectType)
#endif

#if RELEASE
	#if USE_RTOS
	#else
		#define kObjectSize 6	// words (>including< struct bits)
	#endif
#else
	#if USE_RTOS
	#else
		#define kObjectSize 6	// words (>including< struct bits)
	#endif
#endif

// OBJECT is an ('the') abstract base class - it is only inherited from (not strictly enforced)

#define OBJECT_MAGIC 0x2f4f // '/O'

// ObjectInfo can be used to obtain details of an object and create a new object
typedef struct ObjectInfo
	{
	word        magic;
	ObjectClass type;
	const char *name;
	ObjectSize  size;
	// any bigger overlaps system usage...
	} ObjectInfo;
#define __INFO_MAGIC 0x2f6f // '/o'

typedef PTR OBJECT_ELEMENT;
#define DefineObject(class, type, size, name) \
	OBJECT_ELEMENT name##Object[size]=         \
		{                                       \
		(OBJECT_ELEMENT) __INFO_MAGIC,          \
		(OBJECT_ELEMENT) type,                  \
		(OBJECT_ELEMENT) #name,                 \
		(OBJECT_ELEMENT) size,                  \
		};\
	__const_object__ class name = (class) &name##Object
#define ExternObject(class, name) extern const class name

///////////////////////////////////////////////////////////////////////////////////////
//
// object inheritance info
//
///////////////////////////////////////////////////////////////////////////////////////

#define OBJECT_HEADER \
	/* private */ word __object[kObjectSize - 2]; /* # words excluding publics */    \
	/* public  */ byte index;                     /* most recently invoked method */ \
	/* public  */ byte stack;                     /* reserved - do not use! */       \
	/*         */
#define OBJECT_BITS   \
	/* private */ word _object_bits :5; \
	/* public  */ word isvirtual    :1; \
	/* public  */ word enabled      :1; \
	/* public  */ word trace        :1; \
	/*         */

// object API indices
enum
	{
	__objectCreate = 0,
	__objectDelete,
	__objectEncode,
	__objectDecode,

	__objectMaxApiIndex,
	};

// object alert entries
enum
	{
	kObjectErrorBadMagic = 1,			//	{kFatal, "object has bad magic marker"}
	kObjectErrorUnknownClass,			//	{kError, "%s - unknown class"}
	kObjectErrorInvalidSize,			//	{kError, "invalid - size for '%s'"}
	kObjectErrorNotFound,				//	{kError, "not found in %s table"}
	kObjectErrorInvalidUse,				//	{kError, "invalid use of %s class"}
	kObjectErrorVirtualClass,			//	{kError, "invalid use of virtual class %s"}
	kObjectErrorInvalidMethod,			//	{kFatal, "invalid method invocation (object initialized?)"}
	kObjectErrorMethodUndefined,		//	{kError, "method is not defined for the object"}
	kObjectErrorObjectDisabed,			//	{kDebug, "ignoring disabled wait/post"}
	kObjectErrorAlreadyInited,			//	{kDebug, "object has already been created"}
	kObjectErrorMethodDepreciated,	//	{kDebug, "method has been depreciated"}
	kObjectErrorMethodUnsupported,	//	{kDebug, "method is unsupported this version"}
	//
	kMaxObjectError
	};

///////////////////////////////////////////////////////////////////////////////////////
//
// object API
//
///////////////////////////////////////////////////////////////////////////////////////

#define getMethod(object, method) ((METHOD) ((METHOD)_sys->findMethod)(object, method))
#define invokeMethod(obj, index, args) getMethod(obj, index) args
#define invokeInherited(obj, args) \
	do{ SEGPTR(object, obj)->stack++; getMethod(obj, SEGPTR(object, obj)->index) args; SEGPTR(object, obj)->stack--; }while(0)

//
// object vtable methods
//
STATUS objectCreate(OBJECT, ...);          // make a new object
STATUS objectDelete(OBJECT);               // delete an object

STATUS objectEncode(OBJECT);               // encode an object (from internal form)
STATUS objectDecode(OBJECT);               // decode an object (into internal form)

//
// object static methods
//
STATUS objectError(OBJECT obj, const char *method, ObjectClass type, word errno, ...);
                                                       // output an error message
STATUS objectEnable(OBJECT, bool);         //
STATUS objectTrace(OBJECT, bool);          //

ObjectInfo objectInfo(OBJECT obj);         // get object info
ObjectName objectName(OBJECT obj);         // get object name
ObjectType objectType(OBJECT obj);         // get object type
ObjectSize objectSize(OBJECT obj);         // get object size
CLASS objectClass(OBJECT obj);             // fetch the class object of an object

//OBJECT getObjectByName(OBJECT, const char *);        // get object by name
//OBJECT setObjectByName(OBJECT, const char *);        // set object by name tuple

CLASS objectPrintf(OBJECT obj);            // encodes an object to STDOUT

#if 1
// test function
STATUS objectTest(word, word, word, word);
#endif
#endif//__OBJECT_H

