/*
 * File: Koliada - assert.h
 *
 * Contains: definition of assert
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Copyright:	 1989-2012 by Guy McIlroy
 *  All rights reserved.
 *
 *  This file contains confidential, unpublished, proprietary source code.
 *  The copyright notice above does not evidence any actual or intended
 *  publication of such source code.
 *
 *  When deploying, or explicitly linking to, this file or any derivitive,
 *  you may do so under the terms of GPL2 for all non commercial deployments.
 *
 *  When implicitly linking to this file or any derivitive, via Koliada dynamic
 *  linking, your application does not fall under the terms of this license unless
 *  that application also explicitly links to this file or its derivitives.
 *
 *  You must leave all copyright notices and output in place.
 *
 *  You may add copyright notices & output as appropriate.
 *
 *  For commercial licensing contact www.koliada.com
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */
#ifndef __ASSERT_H
#define __ASSERT_H
/*
--------------------------------------------------------------------
http://www.opengroup.org/onlinepubs/009695399/basedefs/assert.h.html
NAME

    assert.h - verify program assertion

SYNOPSIS

    #include <assert.h>

DESCRIPTION

    [CX] [Option Start] The functionality described on this reference page is aligned with the ISO C standard. Any conflict between the requirements described here and the ISO C standard is unintentional. This volume of IEEE Std 1003.1-2001 defers to the ISO C standard. [Option End]

    The <assert.h> header shall define the assert() macro. It refers to the macro NDEBUG which is not defined in the header. If NDEBUG is defined as a macro name before the inclusion of this header, the assert() macro shall be defined simply as:

    #define assert(ignore)((void) 0)

    Otherwise, the macro behaves as described in assert().

    The assert() macro shall be redefined according to the current state of NDEBUG each time <assert.h> is included.

    The assert() macro shall be implemented as a macro, not as a function. If the macro definition is suppressed in order to access an actual function, the behavior is undefined.

-------------------------------------------------------------------
http://www.opengroup.org/onlinepubs/009695399/functions/assert.html
NAME

    assert - insert program diagnostics

SYNOPSIS

    #include <assert.h>

    void assert(scalar expression);

DESCRIPTION

    [CX] [Option Start] The functionality described on this reference page is aligned with the ISO C standard. Any conflict between the requirements described here and the ISO C standard is unintentional. This volume of IEEE Std 1003.1-2001 defers to the ISO C standard. [Option End]

    The assert() macro shall insert diagnostics into programs; it shall expand to a void expression. When it is executed, if expression (which shall have a scalar type) is false (that is, compares equal to 0), assert() shall write information about the particular call that failed on stderr and shall call abort().

    The information written about the call that failed shall include the text of the argument, the name of the source file, the source file line number, and the name of the enclosing function; the latter are, respectively, the values of the preprocessing macros __FILE__ and __LINE__ and of the identifier __func__.

    Forcing a definition of the name NDEBUG, either from the compiler command line or with the preprocessor control statement #define NDEBUG ahead of the #include <assert.h> statement, shall stop assertions from being compiled into the program.

RETURN VALUE

    The assert() macro shall not return a value.

ERRORS

    No errors are defined.
*/
#ifndef assert
	// overide this with assert for your system
	#ifdef NDEBUG
		#define assert(ignored)((void) 0)
	#else
		#define useAssertFailed 1
		extern void (*_assertFailed)(char *, unsigned, char *);
		#define assert(e) ((e) ? (void)0 : _assertFailed(__FILE__, __LINE__, #e))
		#if useAssertFailed
			void (*setAssert(void (*fn)(char *name, unsigned line, char *exp)))(char *name, unsigned line, char *exp);
		#endif
	#endif
#endif
#endif //__ASSERT_H
