/*
 *	File: Koliada - rtos.h
 *
 *	Contains: Koliada task library support
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Copyright:	 1989-2012 by Guy McIlroy
 *  All rights reserved.
 *
 *  This module contains confidential, unpublished, proprietary source code.
 *  The copyright notice above does not evidence any actual or intended
 *  publication of such source code.
 *
 *  This code may only be used under licence from Koliada, LLC - www.koliada.com
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */
#ifndef __TASKS_H
#define __TASKS_H

#if USE_TASK_MANAGER
	// tasking is managed by a Task Manager component
	// simply include its header
	#include TASK_MANAGER_H
#else
	// !USE_TASK_MANAGER
	#include "Koliada.h"

	// tasking in system without shared components can be;
	// threaded (stack based)
	//    - needs sufficient memory to provide one stack per task
	//    - may be pre-emtive or non pre-emptive
	// or non-threaded (co-routines)
	//    - uses same stack for all tasks
	//    - can only be non pre-emtive
	//    - can only 'yield' from the task function (uses 'Duffs Device')

	#if USE_THREADS
		// stack based task handing
		// task stacks are statically assigned (fixme!!)

		#define TaskBegin(fn, ssize, args)\
			TASK fn##Task;\
			byte fn##Stack[ssize];\
		   static void fn##Thread args\
		   {/* TRACE_TASK((">>%s\n", #fn)); */{\
		   T_Enter();\

		#define TaskEnd(fn)\
			goto fn##Done;\
			fn##Done: T_Leave();\
			}/*TRACE_TASK(("<<%s\n", #fn));*/}

	#else
		// !USE_THREADS
		// coroutine task handing, for ultra small systems
		extern const exception EX_SWITCH;

		// no task stacks, but we keep the macro invocation symetric
		// the same code can switch between different systems

		#define TaskBegin(fn, ssize, args)\
			TASK fn##Task;\
		   static void fn##Thread args\
		   {/* TRACE_TASK((">>%s\n", #fn)); */{\
		   mpxLive();\
		   /*{do{*/switch(mpxGetLine()){case 0: do {

		#define TaskDone(fn)	goto fn##Done;

		#define TaskEnd(fn)\
			} while (0);}\
			goto fn##Done;\
			fn##Done:\
				/* mpxSetLine(0);/**/\
				mpxExit();\
			}/*TRACE_TASK(("<<%s\n", #fn));*/}

	#if 0
		#define CATCH_TASK \
			/* store the task context so we can re-start the task */         \
			mpxStore((word) tp);	mpxSetLine(__LINE__); Punt(); case __LINE__:\
			tp = (CM_TIMER_T *) mpxFetch();                                  \
			exc_begin(&ExFrame);	/* rebegin the exception frame */
	#endif

	#endif// !USE_THREADS
	/* depreciated */ #define BeginThread TaskBegin
	/* depreciated */ #define   EndThread TaskEnd

	typedef enum
		{
		kIntrTask = 0,	// interrupt threads ONLY!!
		kTrapTask,		// real-time 'trap' threads (non-preempted)
		kSnapTask,		// utility threads (eg: signals & servers)
		kFastTask,		// fast user threads
		kUserTask,		// normal user threads
		kSlowTask,		// slow (back-ground) threads
		kIdleTask,		// idle threads (only run if no other threads)
		kDeadTask,		// no longer in service (never rescheduled)

		kLastTask,		// not used!!
		} TaskClass;
	#define kMaxTaskClass		kLastTask

#ifndef kDefaultNameSize
	#define kDefaultNameSize	16
#endif
#ifndef kTaskNameSize
	#define kTaskNameSize		kDefaultNameSize
#endif
#if kSemNameSize
	#define kMuxNameSize	kSemNameSize
#else
	#define kSemNameSize			kDefaultNameSize
#endif
#if kMuxNameSize
	#define kSemNameSize	kMuxNameSize
#else
	#define kMuxNameSize			kDefaultNameSize
#endif

#ifndef kDefaultTableSize
	#define kDefaultTableSize	16
#endif
#ifndef kMaxTcbTableSize
	#define kMaxTcbTableSize	kDefaultTableSize
#endif
#ifndef kMaxMuxTableSize
	#define kMaxMuxTableSize	kDefaultTableSize
#endif
#ifndef kMaxSemTableSize
	#define kMaxSemTableSize	kDefaultTableSize
#endif

#if 0	// depreciated
	#define TCB_CLASS_INTR		kIntrTask
	#define TCB_CLASS_TRAP		kTrapTask
	#define TCB_CLASS_SNAP		kSnapTask
	#define TCB_CLASS_FAST		kFastTask
	#define TCB_CLASS_USER		kUserTask
	#define TCB_CLASS_SLOW		kSlowTask
	#define TCB_CLASS_IDLE		kIdleTask
	#define TCB_MAXCLASS			kMaxTaskClass
	#if 0
	#define TCB_CLASS_YIELD		7					/* pass over these threads (yielded) */
	#define TCB_CLASS_DEAD		8					/* suspended threads (never re-scheduled) */
	#endif
#endif

	#define MPX_MESSAGE_HEADER\
		TASK pid;						/* sender */\
		word type;						/* type */
	//	char *name;						/* name for debug purposes */
	//	char name[MAXSTRUCTNAME];	/* name for debug purposes */

	typedef struct MPX_MESSAGE		/* Koliada base message class */
		{
		MPX_MESSAGE_HEADER
		} MPX_MESSAGE;

	#ifndef ANY
		#define ANY ((OBJECT)-1)
	#endif

	///////////
	//
	//
	//
	///////////

#ifndef __MPX_H	// not inside the MPX library
	extern OBJECT *____mpx;
	#define CURPID ((TASK)  ____mpx[0x0])
	#define CURSEG ((HANDLE)____mpx[0x1])

	// Koliada RTOS API
	word mpxTick(void);
	word mpxLive(void);
	word mpxExit(void);
	word mpxYield(void);
	word mpxReady(TASK);
	word mpxBlock(TASK);
	word mpxMuxWait(SEMAPHORE);
	word mpxMuxPost(SEMAPHORE);
	word mpxSemWait(SEMAPHORE);
	word mpxSemPost(SEMAPHORE);
	word mpxMsgWait(TASK, void*, word);
	word mpxMsgPost(TASK, void*, word);
	word mpxAstPost(TASK, TaskClass, word, AST_FN, word, ...);
	word mpxEventWait(EVENT, word);
	word mpxEventPost(EVENT, word);
	word mpxRaiseSignal(SIGNAL, word);
	void mpxPostInterrupt(TASK);

	// the API is abstracted via macros
	// macros abstract out implementation variations
	#if USE_THREADS
		#define T_Tick()							mpxTick()
		#define T_Yield()							mpxYield()
		#define T_Ready(pid)						mpxReady(pid)
		#define T_Block(pid)						mpxBlock(pid)
		#define WaitMux(sem)						mpxMuxWait(sem)
		#define PostMux(sem)						mpxMuxPost(sem)
		#define WaitSem(sem)						mpxSemWait(sem)
		#define PostSem(sem)						mpxSemPost(sem)
		#define WaitMsg(pid, ptr, size)		mpxMsgWait(pid, ptr, size)
		#define PostMsg(pid, ptr, size)		mpxMsgPost(pid, ptr, size)
		#if COMMA_ELLIPSIS
			#define PostAst(task, class, p, fn, args, ...)\
														mpxAstPost(task, class, p, fn, args)
		#else
			#define PostAst(task, class, p, fn, args...)\
														mpxAstPost(task, class, p, fn, args)
		#endif
	#if 0
		#define WaitEvent(e, ret)				mpxEventWait(e, ret)
		#define PostEvent(e, ret)				mpxEventPost(e, ret)
		#define PostInterrupt(pid)				mpxPostInterrupt(pid)
		#define RaiseSignal(sig, val)			mpxRaiseSignal(sig, val)
	#endif

	#else
		// !USE_THREADS
		word mpxGetLine();
		word mpxSetLine(word line);
		word mpxStore(OBJECT);
		OBJECT mpxFetch();

		extern word inIntHandler;
		#define I_Enter()							(++inIntHandler)
		#define I_Leave()							(inIntHandler--)

//		#define mpxCase(line)					// normally does nothing

		#define T_Tick()							mpxTick()
		#define T_Live()							mpxLive()
		#define T_Exit()							mpxExit()
		#define T_Yield()							mpxYield()
		#define T_Switch()						mpxSetLine(__LINE__); Raise(EX_SWITCH); case __LINE__:
		#define T_Ready(pid)						mpxReady(pid)
		#define T_Block(pid)						mpxSetLine(__LINE__); mpxBlock(pid); case __LINE__:
		#define WaitMux(mux)						mpxMuxWait(mux)
		#define PostMux(mux)						mpxMuxPost(mux)
		#define WaitSem(sem)						mpxSemWait(sem)
		#define PostSem(sem)						mpxSemPost(sem)
		#define WaitMsg(pid, ptr, size)		mpxMsgWait(pid, ptr, size)
		#define PostMsg(pid, ptr, size)		mpxMsgPost(pid, ptr, size)
		#if COMMA_ELLIPSIS
			#define PostAst(task, class, p, fn, args, ...)\
														mpxAstPost(task, class, p, fn, args)
		#else
			#define PostAst(task, class, p, fn, args...)\
														mpxAstPost(task, class, p, fn, args)
		#endif
	#if 0
		#define WaitEvent(e, ret)				mpxEventWait(e, ret)
		#define PostEvent(e, ret)				mpxEventPost(e, ret)
		#define PostInterrupt(pid)				mpxPostInterrupt(pid)
		#define RaiseSignal(sig, val)			mpxRaiseSignal(sig, val)
	#endif

#if 0
#warning "Fix these!"
#define mpxGetTaskName(pid, task)
#define mpxShowTasks()
#endif
#endif// USE_THREADS

	// common MPX library entry points
	STATUS mpxCreateThread(TASK*, char *name, THREAD, TaskClass, OBJECT ss, OBJECT sp, word argc, ...);
	STATUS mpxDeleteThread(TASK);

	STATUS mpxCreateSem(SEMAPHORE*, word value, char* name);
	STATUS mpxCreateMux(SEMAPHORE*, char* name);
	STATUS mpxDeleteSem(SEMAPHORE);
	STATUS mpxDeleteMux(SEMAPHORE);
	STATUS mpxGetSemValue(SEMAPHORE, word *);

	// some handy function mappings
#if 0
	#if COMMA_ELLIPSIS
		#define CreateTask(task, name, thread, class, ss, sp, argc, ...)\
														mpxCreateThread(&task, name, thread, class, ss, sp, argc)
	#else
		#define CreateTask(task, name, thread, class, ss, sp, argc...)\
														mpxCreateThread(&task, name, thread, class, ss, sp, argc)
	#endif
#endif
	#define DeleteTask(task)					mpxDeleteThread(task)
	#define CreateSem(sem, value, name)		mpxCreateSem(&sem, value, name)
	#define DeleteSem(sem)						mpxDeleteSem(sem)
	#define CreateMux(mux, name)				mpxCreateMux(&mux, name)
	#define DeleteMux(mux)						mpxDeleteMux(mux)
	#define GetSemValue(sem, value)			mpxGetSemValue(sem, &value)
	#define TimeSlice(task)						mpxTimeSlice(task)

	// ease of use - careful of the need for a trailing arg list!
	// for eample: createTask(Task, kTrap, 2, arg1, arg2);
	#if USE_THREADS
		#if COMMA_ELLIPSIS
			#define createTask(name, class, args, ...)\
				mpxCreateThread(&name##Task, #name, name##Thread, class,(OBJECT)name##Stack, sizeof(name##Stack), args)
		#else
			#define createTask(name, class, args...)\
				mpxCreateThread(&name##Task, #name, name##Thread, class,(OBJECT)name##Stack, sizeof(name##Stack), args)
		#endif
	#else
		#define createTask(name, class, args)\
			mpxCreateThread(&name##Task, #name, name##Thread, class, 0, 0, 0)
	#endif
	#define deleteThread(name)			mpxDeleteThread(name##Task)

	#define deleteTask(name)			mpxDeleteThread(name##Task)
	#define createSem(sem, val)		mpxCreateSem(&sem, val, #sem)
	#define createMux(mux)				mpxCreateMux(&mux, #mux)
	#define deleteSem(sem)				DeleteSem(sem)
	#define deleteMux(mux)				DeleteMux(mux)

#endif //__MPX_H

#if USE_MPX_EXTENDED
	// extended MPX library functions
	STATUS mpxSetDebugMask(BitField mask);
	STATUS mpxSetThreadSeg(TASK, SEGMENT);
	STATUS mpxSetPriority(TASK pid, word priority);
	STATUS mpxNewObject(ObjectType, size_t size, OBJECT *seg);
	STATUS mpxDelObject(OBJECT seg);
	STATUS mpxGetTaskName(TASK id, char *name);

	// mailbox functions
	STATUS mpxCreateMailbox(MAILBOX *, WaitMode, char *);
	STATUS mpxDeleteMailbox(MAILBOX);
	STATUS mpxPostWaitMsg(MAILBOX, MAILMSG);
	STATUS mpxPostAsyncMsg(MAILBOX, MAILMSG, WaitMode);
	STATUS mpxWaitAsyncMsg(MAILBOX, MAILMSG, WaitMode);

	#define CreateMailbox(b, mode)		mpxCreateMailbox(&b, mode, #b)
	#define DeleteMailbox(b)				mpxDeleteMailbox(b)
#endif

#endif// !USE_TASK_MANAGER
#endif//__TASKS_H
