/*
 *	File: Koliada - gpio.h
 *
 *	Contains: system gpio port descriptor definitions
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Copyright  1989-2012 by Guy McIlroy
 *  All rights reserved.
 *
 *  This module contains confidential, unpublished, proprietary source code.
 *  The copyright notice above does not evidence any actual or intended
 *  publication of such source code.
 *
 *  This code may only be used under licence from Koliada, LLC - www.koliada.com
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */
#ifndef __GPIO_H
#define __GPIO_H
/*
	CCXXXX devices (generally) have three digital I/O ports implemented, P0, P1, P2. P0 & P1
	have eight I/O pins, P2 has 4. Every I/O pin is individually configurable for input or
	output direction, and each I/O line can be individually read from or written to.

	SFRs are available both from direct access memory and extended 'unified' memory.
	Since we access all i/o as memory pointers, we can only use the extended memory

	P0 and P2 have special case IEN
*/
#ifndef __TARGET_H
	extern SystemMemory sysData;
	#define _sys (&sysData)
	#define GPIO	(*((GpioTBL *)_sys->gpio))
#endif

// GPIO port descriptor
typedef struct PortDescriptor
	{
	UInt16 io;	// input / output
	UInt16 dir;	// direction
	UInt16 sel;	// function selection
	UInt16 ren;	// pullup/down resistor enable - corresponding bit in PxDIR selects pullup/down
	UInt16 ifg;	// port interrupt flag
	UInt16 ier;	// interrupt enable register
	UInt16 irq; // IRQ vector
	} PortDescriptor;

#define PORT_IRQ(port)   __BoardCFG->ports.portTable[port].irq

typedef struct PortConfig
	{
	byte count; 					// number of gpio ports
	PortDescriptor *portTable;	// gpio config data - worst case this is a 50 byte table
	} PortConfig;

// GPIO pin descriptors (these are inited in initGpio)
typedef struct PinDescriptor
	{
	word port;	// pin port
	word mask;	// pin mask
	struct
		{
		word dir : 1;		// direction:        0 - input,   1 - output
		word ies : 1;		// edge select:      0 - rising,  1 - falling
		word ren : 1;		// pullup/dn:        0 - pull,    1 - 3-state
		word inv : 1;		// signal invert:    0 - normal,  1 - invert
		word set : 1;		// initial value:    0 - low,     1 - high
		word ien : 1;		// interrupt enable: 0 - disable, 1 - enable
		};
	} PinDescriptor;
// single descriptor can describe multiple pins with same port/config

// pin descriptor macros
#define PIN_PORT(pin) (__BoardCFG->pins.pinTable[pin].port)
#define PIN_MASK(pin) (__BoardCFG->pins.pinTable[pin].mask)

#define PIN_DIR(pin) (__BoardCFG->pins.pinTable[pin].dir)
#define PIN_IES(pin) (__BoardCFG->pins.pinTable[pin].ies)
#define PIN_REN(pin) (__BoardCFG->pins.pinTable[pin].ren)
#define PIN_INV(pin) (__BoardCFG->pins.pinTable[pin].inv)
#define PIN_SET(pin) (__BoardCFG->pins.pinTable[pin].set)
#define PIN_IEN(pin) (__BoardCFG->pins.pinTable[pin].ien)

typedef struct PinConfig
	{
	byte count; 						// number of gpio pins
	PinDescriptor *pinTable;	// pin config data
	} PinConfig;

// A pin mapping table allows different gpio pins to be mapped to the same symbolic constant
// this allows binary code compatibility across multiple targets with differing pin assingments

// gpio MAP descriptor
typedef struct MapDescriptor
	{
	word pin;						// pin table index
	} MapDescriptor;

typedef struct MapConfig
	{
	word maps;						// number of mapped pins
	MapDescriptor *mapTable;	// pin map config data
	} MapConfig;

// gpio pin table indexes
#define GPIOX  ((word)-1)	// undefined
#define GPIO1  0
#define GPIO2  1
#define GPIO3  2

// CC8051 synonyms for GPIO ports
#define PORT0 0
#define PORT1 1
#define PORT2 2

// port address macros
#define PIN_ADDR(pin)  ((PTR) (word)__BoardCFG->ports.portTable[PIN_PORT(pin)].io)
#define POUT_ADDR(pin) ((PTR) (word)__BoardCFG->ports.portTable[PIN_PORT(pin)].io)

// port access macros
#define PIN(p)		*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].io))
#define POUT(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].io))
#define PDIR(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].dir))
#define PSEL(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].sel))
#define PREN(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].ren))
#define PIFG(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].ifg))
#define PIER(p)	*((volatile unsigned char *)((word)__BoardCFG->ports.portTable[p].ier))

// GpioTBL	                   		`           // gpio function table
typedef CONST struct GpioTBL
	{
	word (*GET)(word pin);				// GET pin state
	word (*SET)(word pin, word set);	// SET pin state
	word (*XOR)(word pin);				// XOR pin state

	word (*DIR)(word pin, word set);	// set pin DIRection
	word (*SEL)(word pin, word set);	// SELect pin function
	word (*REN)(word pin, word set);	// Resistor pullup ENable (pullup/pulldn/none)

	word (*INT)(word pin);				// get pin INTerrupt flag state
	word (*IFG)(word pin, word set);	// set pin Interrupt FlaG
	word (*IES)(word pin, word set);	// set pin Interrupt Edge Select (rising/falling)
	word (*IEN)(word pin, word set);	// set pin Enterrupt ENable
	} GpioTBL;

// GpioTBL is not extensible, but it is replacable via _sys->gpio
// replacements must match this vtable!

#endif//__GPIO_H
